;*=====================================================================*/
;*    Datalog high-level API					       */
;*=====================================================================*/
; Creation    : Mon 22 Mar 2004 04:10:03 PM MSK
; Last change : Fri 21 May 2004 11:36:31 AM MSD
; version1.2.0 
;
; Exported: 
;
; dtlog-edb
; 
; 
; dtlog-eval
; dtlog-eval-query
; dtlog-print-result
; 
;=============================================================================
(module dtlog-lib
    (library sqlite)
    (library srfi-1)
    (import os)
    (include "../common/myenv.scm")
    (include "../common/bigloo/common-bigloo.scm")
    (include "../src/bench-tools.scm")
    (include "../src/rel-algebra.scm")
    (include "../src/unify.scm")
    (include "../src/qsqr.scm")
    (include "../src/calculating.scm")
    (include "../src/dtlog-to-ra.scm")
    (include "../src/dtlog-parse.scm")
 (export
   (dtlog-edb filename) 
;   (dtlog-rule rule)
;   (dtlog-query query)
   (dtlog-eval rules edb-in)
   (dtlog-eval-query rules query edb-in)
   (dtlog-print-result idb-in)
   
   (my-bind names vals name)
   (parse-identifier identifier)
   (parse-number number)
   (parse-operation op expr1 expr2)
   (parse-assign identifier expr)
   (parse-aggregate id1 id2 predicate-list aggr-function)

  )
 (eval    (export-all))
)
;=============================================================================
; Procedures

; dtlog-edb :: filename -> db
(define (dtlog-edb filename) 
  filename)

; dtlog-eval :: rules -> db -> idb
; dtlog-eval apply rules to edb, returns result-idb. Evaluation is performed with common-semi-immediate method
; rules is a list of values are taken with dtlog-rule, 
; edb is a value is taken with dtlog-edb
; result-idb should be process with dtlog-print-result
(define (dtlog-eval rules edb-in) 
  (define filename edb-in)
  (define idb-names (get-idb-names rules))
  (receive (edb-names idb edb) (read-idb-edb filename idb-names) 
    (let ((ra-eqs (dtlog->ra rules)))
      (values idb-names (common-semi-immediate ra-eqs idb-names idb edb-names edb)))))

; dtlog-eval-query :: rules -> query -> edb -> result
; dtlog-eval-query apply qsq method to edb with rules and query as a program
; rules is a list of values are taken with dtlog-rule, 
; query is a result of dtlog-query method
; result is a list of facts (tuples = vectors)
(define (dtlog-eval-query rules query edb-in)
  (define filename edb-in)
  (define idb-names (get-idb-names rules))  
  (receive (edb-names idb edb) (read-idb-edb filename idb-names) 
    (let* ((new-query (replace-avariable query))
	  (ra-eqs (dtlog->ra rules))
	  (idb (common-semi-immediate ra-eqs idb-names idb edb-names edb))
	  (q-res ((right-part-expr new-query #f) (append idb edb) (append idb-names edb-names))))
;    (cout idb nl)	  
    q-res
)))


;dtlog-print-result :: idb -> unspecified
(define (dtlog-print-result idb-in) 
  (receive (idb-names idb) idb-in
    (output-res idb-names idb "Result: ")))

