;; Calculating methods of Datalog
;; version1.2.0 
; Fri 02 Apr 2004 11:03:47 AM MSD
;  
; Relational equation (rule) is represented as s-expression.

; Calculating method Yakobi and Gauss-Zeidel solve a system of relational equations.
; Nonformal description of these methods:

; 1) Math notation
; calculating method :: RelEqs  -> LFP
; RelEqs is a system of relational equations:
; Ri = Ei(R1, R2, ... , Rn) , i = 1..n, Ri are relations and Ei are <relational expression>s.
; Each relation is a set of facts having equal arity.
; Fields of each facts are constants.
; Left Fixed Point (LFP) of RelEqs is a set of relations R1, ..., Rn, applying E1, ..., En to 
; which yield the same result.
;
; 2) Programming notation
; calculating method :: rules -> IDB-names ->  DB -> IDB-values
; calculating method solves a system of relational equations:
; <1st element of IDB-names> = <1st element of rules>
;              ...
; <nth element of IDB-names> = <nth element of rules>
;
; Rules is a list of <rule>'s (see rel-algebra.scm)
; IDB-names is a list of names of relations
; DB = EDB + IDB
; IDB - here is a start IDB-values
; DB is a list of pairs. Each pair consists of the name of relation and 
; immediately relation. 
; Each relation is a list of facts
; Each fact is a vector of constants.
; 
; calculating method returns values of IDB, IDB-values is a list of relations.
;

; Exported:
;
; output-res
; Yakobi
; Yakobi-map (Yakobi in the map terms)
; Gauss-Zeidel
; semi-immediate (only for linear equations)
; common-semi-immediate (for all type of equations)
;
;==============================================================================
; Assistant procedures


; displays message and values of variables from list var-names, which are contained
; in list var-vals

(define (output-res var-names var-vals message)
  (cout message nl)
  (let repeat ((var-names var-names) (var-vals var-vals))
  (if (null? var-names)
      (cout nl)
      (begin
        (cout (car var-names) " = " (car var-vals) nl)
        (repeat (cdr var-names) (cdr var-vals))))))

;(output-res '(%par) '( (#(b a) #(b g) #(a d) #(a e) #(d f) #(c h)) ) "Example: ")
;                          =>    Example: %par = (#(b a) #(b g) #(a d) #(a e) #(d f) #(c h))


;==============================================================================
; Calculation methods of Datalog
; Yakobi method
(define (Yakobi rules idb-names start-idb edb-names edb)
  (define db-names (append idb-names edb-names))
  (define idb-len (length idb-names))
  (let repeat ((idb-vals start-idb))
    (let ((new-idb-vals
           (let iter ((new-idb-vals '()) (old-idb-vals idb-vals) (rules rules) (i 0))
              (if (= i idb-len)
                 new-idb-vals
                 (let* ((expr (car rules))
                        (new-val (expr (append old-idb-vals 
			                     edb) db-names)))
;		   (cout new-val nl)
                   (iter (append new-idb-vals (list new-val))
                         old-idb-vals
                         (cdr rules) (+ i 1)))))))
         (if (list= eq-set? new-idb-vals idb-vals)
             idb-vals
             (begin
;                (output-res idb-names new-idb-vals "Debug: ")
             (repeat new-idb-vals)
	     )))))

; Gauss-Zeidel method
(define (Gauss-Zeidel rules idb-names start-idb edb-names edb)
  (define db-names (append idb-names edb-names))
  (define idb-len (length idb-names))
  (let repeat ((idb-vals start-idb))
    (let ((new-idb-vals
           (let iter ((new-idb-vals '()) (old-idb-vals idb-vals) (rules rules) (i 0))
              (if (= i idb-len)
                 new-idb-vals
                 (let* ((expr (car rules))
			; last append at next string is single difference between G-Zeidel and Yakobi
                        (new-val (expr (append (append! new-idb-vals old-idb-vals) 
;                         (new-val (expr (append old-idb-vals
		                     edb
) db-names)))
			(take! new-idb-vals i)
                   (iter (append new-idb-vals (list new-val))
                         (cdr old-idb-vals
)
;			 old-idb-vals
                         (cdr rules) (+ i 1)))))))
      (if (list= eq-set? new-idb-vals idb-vals)
          idb-vals
          (begin
;           (output-res var-names new-idb-vals "Debug: ")
          (repeat new-idb-vals))))))


; semi-immediate method
(define (common-semi-immediate rules idb-names start-idb edb-names edb)
  (define db-names (append idb-names edb-names))
  (let repeat ((idb-vals start-idb)
               (idb-difs start-idb))
    (let* ((new-idb-difs
           (let iter ((new-idb-difs '()) (old-idb-difs idb-difs) (rules rules) (count 0))
;        (cout "new-idb-difs " new-idb-difs nl)
             (if (null? rules)
                 new-idb-difs
                 (let* ((expr (car rules))
                        (new-dif (difference (expr (append  (append (take idb-vals count)
	                                                                  (list (car old-idb-difs))
                                                                          (drop idb-vals (+ count 1)))
              			                                  edb) db-names)
                        			         (list-ref idb-vals count))))	
;	 	   (cout "new-dif: " nl)
                   (iter (append new-idb-difs (list new-dif))
                         (cdr old-idb-difs)
                         (cdr rules)
                         (+ count 1)))) ))
           (new-idb-vals (map union new-idb-difs idb-vals)))
      (if (list= eq-set? new-idb-difs (make-list (length idb-names) '()))
          (begin 
;  	   (cout "Result step" nl)
;            (output-res idb-names new-idb-difs "Debug difs: ")
            new-idb-vals)
          (begin
;	   (cout "Step" nl)
;           (output-res idb-names new-idb-difs "Debug difs: ")
;           (output-res idb-names new-idb-vals "Debug vals: ")
          (repeat new-idb-vals new-idb-difs))))))



;==============================================================================
; Examples

;Example from Cheri

;(define %par '( #(b a) #(b g) #(a d) #(a e) #(d f) #(c h)))
;(define var-names '(%anc))
;(define rules '( (union (compose %anc %par) %par)
;                 )
;  )
;(define edb (list (list '%par `,%par)))


;(output-res var-names (Yakobi rules var-names edb) "Result Yakobi: ")
;(output-res var-names (Gauss-Zeidel rules var-names edb) "Result Gauss-Z: ")
;(output-res var-names (common-semi-immediate rules var-names edb) "Result semi-immediate: ")

; =>  ((#2(c h) #2(d f) #2(a e) #2(a d) #2(b g) #2(b a) #2(a f) #2(b d) #2(b e) #2(b f)))


