; Main file in Datalog system loads and perform program
; version1.2.0 
; Fri 02 Apr 2004 11:04:39 AM MSD
;
; 
;
; Exported: 
; 
; dtlog-exe-name
; dtlog-version
;
; run-dtlog
;

;==============================================================================
; Variables

; the name of Datalog executable
(define dtlog-exe-name "dtlog")

; version of datalog system
(define dtlog-version "1.2.0 alpha")

;==============================================================================
; Procedures

(define	(handler escape proc mes obj)
  (cout "Datalog error! " proc " : " mes " = " obj nl)
  (escape #f))


; run-dtlog :: opt-args -> unspecified
; dtlog-program is a file contains datalog program
; opt-args - the first is db-filename. If it present, this file is used to 
; perform progra, else db-filename catch from dtlog-program 
(define (run-dtlog argv)
  (define ticks (clocks-per-second))
  (define dtlog-program #f)
  (define db-filename #f)
  (define only_read_db_opt #f)
  (define only_syntactic_opt #f)  
  (define b_opt #f)
  (define bottom-up #t)
  (define calc-method common-semi-immediate)
  (define (ccout . args)
    (if (not b_opt) (apply cout args)))
  (args-parse (cdr argv) 
    (section "help")
     ((("-h" "--help")
       (help "Print this help message"))
      (cout "Usage:" nl)
                 (cout "dtlog [options]" nl)
       (args-parse-usage #f))
     ((("-v" "--version")
       (help "Print version info"))
       (cout dtlog-exe-name " version " dtlog-version nl))
    (section "options")
     (("-f" ?file 
       (help "Specify datalog program"))
       (set! dtlog-program file))
     (("-db" ?file 
       (help "Specify SQLite database file, override db in datalog program"))
       (set! db-filename file))
    (section "analizing")        
     (("--syntactic"
       (help "Do only syntactic analysis"))
       (set! only_syntactic_opt #t))
     (("--read-only"
       (help "Do all before start calculating"))
       (set! only_read_db_opt #t))
    (section "calculations") 
     (("-sn"
       (help "Calculate a program with Semi-Naive method (default)"))
       (set! calc-method common-semi-immediate))
     (("-ya"  
       (help "Calculate a program with Yakobi method "))
       (set! calc-method Yakobi))
     (("-gz"  
       (help "Calculate a program with Gauss-Zeidel method "))
       (set! calc-method Gauss-Zeidel))
     (("-qsq"  
       (help "If a program is calculated with bottom-up, all queries is performed with usual technique. This flag force the calculation with QSQ method. If there are no queries, nothing to do "))
       (set! bottom-up #f))
     (section "misc")
      (("-b"
        (help "Benchmark mode. Perform all calculations and print: number of facts in whole edb, number of inferred facts, number of relevant facts (if query specified), fraction of relevant facts of inferred facts (if query specified), elapsed user time"))
	(set! b_opt #t))
    (else (begin (cout "Illegal argument `" else "'" nl)
                 (cout "Usage:" nl)
                 (cout "dtlog-bigloo [options]" nl))
       (args-parse-usage #f)))
   (if (not (or dtlog-program db-filename)) (exit -1))
                        ;(begin (cout dtlog-exe-name ": too few arguments" nl)
;			;	  (cout "Try `" dtlog-exe-name " --help' for more information" nl))
;				   
   (let* ((undiv-program (let ((x (dtlog-parse dtlog-program)))
			   (if x x (exit -1))))
	  (program (fold 
	    (lambda (x val) (let ((tag (car x)) (cntxt (second x))
		(v1 (first val)) (v2 (second val)) (v3 (third val)))
		(cond ((eq? tag rule-tag)  (list (append v1 (list cntxt)) v2 v3))
		      ((eq? tag query-tag) (list v1 (append v2 (list cntxt)) v3))    			
		      ((eq? tag instr-tag) (list v1 v2 (append v3 (list cntxt)))))))
	    '(()()()) undiv-program))
	    (a! (if only_syntactic_opt (begin (cout "Syntactic analysis ok" nl)(exit 0))))
         (db-sqlite (if db-filename db-filename 
			 (let ((db-instr (find (lambda (x) (eq? (car x) db-tag)) (third program))))
			   (if db-instr (second db-instr) (begin (cout "Database filename isn't specify" nl) 
								(exit -1))))))
         (rules (first program))
         (idb-names (get-idb-names rules))
         (ra-eqs (if bottom-up (try (dtlog->ra rules) handler)))
	 )
      (receive (edb-names idb edb) 
        (read-idb-edb db-sqlite idb-names) 
	  (ccout "EDB-names: " edb-names nl "IDB: " idb nl "EDB: " edb nl)
	  (if only_read_db_opt (exit 0))
          (ccout "Start!" nl)
  	  (receive (total utime stime cutime cstime) (times)
            (let* ((idb (if bottom-up (calc-method ra-eqs idb-names idb edb-names edb)))
	           (q-res (if (null? (second program)) #f
                  (map (lambda (query) 
                         (let* ((new-query (replace-avariable query))
			        (res-idb (if bottom-up 
                                                 ((right-part-expr new-query #f) (append idb edb) (append idb-names edb-names))
                                                 (qsqr rules idb-names edb-names edb new-query))))
			   (list res-idb))) (map cdr (second program))))))
              (receive (ntotal nutime nstime ncutime ncstime) (times)
;      	      (cout "q-res = " q-res nl)
	      (if b_opt 
	        (let* ((res-count (facts-count edb))
		      (q-res-count (if q-res (facts-count (car q-res)) #f))
		      (q-part-res (if q-res-count (/ q-res-count res-count) #f)))
		  (cout res-count "\t")
		  (if bottom-up (cout (facts-count idb) "\t"))
		  (cout q-res-count "\t")
		  (cout q-part-res "\t")
	          (cout (/ (- nutime utime) ticks) "\n"))
		(begin
		  (if bottom-up (output-res idb-names idb "Result: "))
		  (if q-res
                    (map (lambda (query res) 
                         (cout query " = " res nl))
			   (map cdr (second program)) q-res)))) 
	       )))))
   (ccout nl)
)

