;; Most Greater Unificator
;; Thu 01 Apr 2004 10:20:10 AM MSD
;; version1.2.0 
;  

; Exported:
;
; eval-expr?
; logic-pred?
; assign-expr?
; assign-ag-expr?
; atom?
; var?
; const?
; 
; const->set
; apply-term-subst
; apply-atom-subst
; replace1-atom-subst
; apply-rule-subst
; replace1-rule-subst
; compose-subst
; mgu
;
; newvar
; newvar-in-rule
; flat-rule-terms
; corr-list
; only-vars
; eq-var-pos?
; rename-subst?
; rename

; eq-name
; eq-result
; internal-preds
;
; value->number
;
;==============================================================================
; Datalog Syntax

; eval-expr? :: rule-element -> boolean
; returns #t, if rule-element is evaluation expression 
; (logic predicate assignment expression), 
; #f otherwise
(define (eval-expr? rule-element)
  (and (list? rule-element) (>= (length rule-element) 2)
       (list? (car rule-element)) 
       (procedure? (cadr rule-element))))
;(cout "Test eval-expr? " (eval-expr? `((X Z) ,eq?)) nl)

; logic-pred? :: rule-element -> boolean
; if rule-element is logic predicate (==, <=, <, >, >=, !=) returns #t, #f otherwise
(define (logic-pred? rule-element)
  (and (list? rule-element) (equal? (length rule-element) 2)
       (list? (car rule-element)) 
       (procedure? (cadr rule-element))))
  
; assign-expr? :: rule-element -> boolean 
; returns #t if rule-element is assign expression
(define (assign-expr? rule-element) 
  (and (eval-expr? rule-element) (>= (length rule-element) 3)
       (var? (third rule-element))))
;(cout "Test assign-expr? " (assign-expr? `((X Z) ,eq? X)) nl)
;(cout "Test assign-expr? " (assign-expr? `((X Z) ,eq?)) nl)

; assign-ag-expr? :: rule-element -> boolean 
; returns #t if rule-element is assign expression with aggregation
(define (assign-ag-expr? rule-element) 
  (and (eval-expr? rule-element) (eq? (length rule-element) 6)
       (var? (third rule-element)) (var? (fourth rule-element))
       
       ))
;(cout "Test assign-ag-expr? " (assign-ag-expr? `((X Z) ,eq?)) nl)
; -> #f

; atom? :: rule-element -> boolean
; if rule-element is atom, atom? returns #t, #f otherwise
(define (atom? rule-element)
  (and (pair? rule-element) (pred? (car rule-element))))

; var? :: term -> boolean
; if term is variable, var? returns #t, #f otherwise
(define (var? term)
  (and (symbol? term) 
       (let ((chr (string-ref (symbol->string term) 0)))
    	 (or (char-upper-case? chr)
	     (eq? chr #\_)))))
; (var? 'A2) 
; -> #t
; (cout (var? _asdf))
; -> #t

; pred? :: term -> boolean
; if term is predicate, var? returns #t, #f otherwise
(define (pred? term)
  (and (symbol? term) 
       (char-lower-case? (string-ref (symbol->string term) 0))))
; (pred? 'sgc) 
; -> #t


; const? :: term -> boolean
; if term is constant, const? returns #t, #f otherwise
; set of constants is constant too
(define (const? term)
  (or (string? term) (pair? term)))
; (const? "abc") 
; -> #t

; query? :: object -> boolean
; if object is query query? returns #t, #f otherwise
(define (query? object)
  (list? object))

; const->set :: x -> list of const
; useful function if we don't know const is a pair or string
(define (const->set x)
  (if (pair? x) x (list x)))

;rule    

;==============================================================================
; Substitutions, Most Greater Unificator


; apply-term-subst :: term -> substitution -> new term
(define (apply-term-subst term subst)
  (if (null? subst)
      term
      (if (equal? term (car (car subst)))
          (car (cdr (car subst)))
	  (apply-term-subst term (cdr subst)))))
;(cout (apply-term-subst 'X '((Y Z) (X Z) (V "a") (W U))))
; -> 'Z

; apply-atom-subst :: atom -> substitution -> new atom
(define (apply-atom-subst atom subst)
  (cons 
    (car atom) 
    (map (lambda (term)
           (apply-term-subst term subst))
         (cdr atom))))
;(cout (apply-atom-subst '(p X Z "a" U) '((Y Z) (X Z) (V "a") (W U))))
; -> '(p Z Z "a" U)

; replace1-atom-subst :: atom -> substitution -> new atom
; replaces only the first occurence of left side of subst to atom
(define (replace1-atom-subst atom subst)
  (cons 
    (car atom) 
    (let loop ((unproc (cdr atom)))
      (if (null? unproc)
          '()
	  (let* ((old-term (car unproc))
	         (new-term (apply-term-subst old-term subst)))
	    (if (equal? old-term new-term)
	        (cons old-term (loop (cdr unproc)))
		(cons new-term (cdr unproc))))))))
;(cout (replace1-atom-subst '(p X Z D) '((X Y))) nl)
; -> (p Y Z D)

; replace1-rule-subst :: rule -> substitution -> new rule
; replaces only the first occurence of left side of subst in rule
(define (replace1-rule-subst rule subst)
  (if (null? rule)
      '()
      (let* ((old-atom (car rule))
             (new-atom (replace1-atom-subst old-atom subst)))
	(if (equal? old-atom new-atom)
	    (cons old-atom (replace1-rule-subst (cdr rule) subst))
	    (cons new-atom (cdr rule))))))
;(cout (replace1-rule-subst '((p X Z D)) '((X Y))) nl)
; -> ((p Y Z D))

; apply-rule-subst :: rule -> substitution -> new rule
(define (apply-rule-subst rule subst)
  (map (lambda (x) (apply-atom-subst x subst)) rule))
;(cout (apply-rule-subst '( (sgc X Y) (par X X1) (sgc X1 Y1) (par Y Y1)) '((X Z))))
; -> '( (sgc Z Y) (par Z X1) (sgc X1 Y1) (par Y Y1))

; compose-subst :: substitution1 -> substitution2 -> new substitution
; compose substitution1 @ substitution2
(define (compose-subst subst1 subst2)
  (append 
   (delete-duplicates 
     (map 
       (lambda (s-pair) 
         (list (car s-pair) (apply-term-subst (car (cdr s-pair)) subst2)))
         subst1))
   (filter 
     (lambda (s-pair)
       (not (any (lambda (x)
		   (if (equal? (car x) (car s-pair)) #t #f)) 
		 subst1)))		 
     (delete-duplicates subst2))))
;(cout (compose-subst '((X Y) (Y U) (Z U)) '((Y "john") (U T) (V T) (X "sam"))))
; -> '((X "john") (Y T) (Z T) (U T) (V T))

; mgu :: atom1 -> atom2 -> unificator 
(define
 (mgu atom1 atom2)
 ;(cout nl atom1 nl atom2 nl nl)
 (if (or (not (equal? (car atom1) (car atom2))) 
         (not (eq? (length atom1) (length atom2))))
     #f
     (let rpt ((terms1 (cdr atom1)) (terms2 (cdr atom2)) (subst '()))
       (if (null? terms1)
	   subst 
           (let* ((term1 (car terms1)) (term2 (car terms2))
	          (new-term1 (apply-term-subst term1 subst))
		  (new-term2 (apply-term-subst term2 subst)))
	     (if (not (equal? new-term1 new-term2))
        	 (if (var? new-term2)
		     (rpt (cdr terms1) (cdr terms2) 
		          (compose-subst subst (list (list new-term2 new-term1))))
		     (if (var? new-term1)
   		         (rpt (cdr terms1) (cdr terms2) 
		           (compose-subst subst (list (list new-term1 new-term2))))
			 #f))))))))

;(cout (mgu '(p X Z "a" U) '(p Y Y V W)))
; -> '((Y Z) (X Z) (V "a") (W U))
;(cout (mgu '(p "a") '(p "b")))
; -> #f

;==============================================================================
; Datalog Syntax Assistant Functions

; newvar :: variables -> new variable
; doesn't exist variable in variables that equals new variable
(define (newvar vars)
;  (cout "newvar: " vars nl)
  (string->symbol (let ((chr #\X) (str "X"))
    (string-append str
      (number->string 
        (+ (apply max 0
	        (filter
                (lambda (x) x)
                (map 
                  string->number
                  (map 
                    (lambda (x) 
                      (list->string (cdr (string->list x))))
                    (filter (lambda (x) (char=? chr (string-ref x 0)))
                      (map symbol->string vars)))))) 1))))))
; (cout (newvar '(X1 Y)))
; -> X2

; flat-rule-terms :: rule -> lst
; returns list of all rule's terms (without preds), without swapping
(define (flat-rule-terms rule)
  (filter (lambda (x) (not (pred? x)))
    (fold (lambda (atom lst) 
            (cond ((assign-expr? atom) (append lst (list (third atom))))
	      ((logic-pred? atom) lst)
	      (else (append lst atom)))) '() rule)))
;(cout "Test1 "(flat-rule-terms '((anc X Y) (anc X Z) (par Z Y))) nl)
; -> '(X Y X Z Z Y)
;(cout "Test2 "(flat-rule-terms `((anc X Y) ((X) ,eq?) (anc X Z) (par Z Y))) nl)
; -> '(X Y X Z Z Y)
;(cout "Test3 " (flat-rule-terms `((anc X Y) ((X) ,eq? W) (anc X Z) (par Z Y))) nl)
; -> '(X Y W X Z Z Y)

; flat-rule-vars :: rule ->lst
(define (flat-rule-vars rule)
  (filter var? (flat-rule-terms rule)))

; newvar-in-rule :: rule -> new variable
; returns new variable that doesn't exist in rule
(define (newvar-in-rule rule)
;  (cout rule nl)
  (newvar (filter var? (flat-rule-terms rule))))
;(cout (newvar-in-rule '((%anc X1 Y) (%anc X Z) (%par Z Y))) nl)
; -> X2

; corr-list :: rule -> list
; analog - rel-algebra.scm.proj-fields
; for each index i of rule's left side vars ai,
;  if exists bj : ai = bj, bj is first such var in right rule's side
;  	bj includes in the result list
;       #f otherwise        
(define (corr-list rule)
  (map (lambda (ai) 
         (list-index (lambda (bj) (equal? ai bj)) (flat-rule-terms (cdr rule)))) 
      (cdr (car rule))))
;(cout (corr-list '((anc X Y) (anc X Z) (par Z Y))) nl)
; -> '(0 3)
;(cout (corr-list '((anc X U Y) (anc X Z) (par Z Y))) nl)
; -> '(0 #f 3)

; only-vars :: atom -> new-atom
; returns new-atom are consisted only with vars
 (define (only-vars atom)
  (cons (car atom) (filter var? (cdr atom))))
;(cout (only-vars '(sgc X "b" Y)) nl)
;(sgc X Y)

; eq-var-pos? :: c-query1 -> c-query2 -> boolean
; return #t if positions of vars and constants are equal, #f otherwise
(define (eq-var-pos? atom1 atom2)
  (fold (lambda (x y val)
          (and val (cond ((and (var? x) (var? y)) #t)
	                 ((and (const? x) (const? y)) #t)
			 (else #f)))) #t (cdr atom1) (cdr atom2)))
;(cout (eq-var-pos? '(par Y ("a")) '(par X "a")) nl nl)

; rename-subst? :: subst -> boolean
; returns #t if subst is only rename substitution, #f otherwise
(define (rename-subst? subst)
  (not (find (lambda (x) (const? (cadr x))) subst)))
;(cout (rename-subst? '((X Y))) nl)

; rename :: common-query -> rule -> new-rule
; all variables in rule that exist in common-query are renamed
(define (rename c-qry rule)
; (cout c-qry nl rule nl nl)
 (apply-rule-subst 
   rule
   (let* ((c-qry-vars (filter var? c-qry))
          (rule-vars (delete-duplicates (filter var? (fold append '() rule)))))
;     (cout c-qry-vars nl rule-vars nl)
     (let rpt ((rule-vars rule-vars) (r-newvars (append rule-vars c-qry-vars)))
       (if (null? rule-vars)
           '()
 	  (if (find (lambda (x) (eq? x (car rule-vars))) c-qry-vars)
 	    (let ((n-var (newvar r-newvars)))
   	      (cons (list (car rule-vars) n-var)
 	            (rpt (cdr rule-vars) (list n-var))))
 	    (rpt (cdr rule-vars) r-newvars)))))))

;(cout (rename '(sgc X Y) '((sgc X Y) (par X X1) (sgc X1 Y1) (par Y Y1))) nl)
; -> ((Y X2) (X X3))
; -> ((sgc X3 X2) (par X3 X1) (sgc X1 Y1) (par X2 Y1))

;==============================================================================
; Internal Datalog predicates

; eq-name is a system notation of Datalog's "eq"
(define eq-name 'eq)

; eq-result :: sub-aim -> binary-relation
; this is implementation of Datalog's "eq". 
; Sub-aim is a unificated eq-predicate (with set of constant)
; If the first or the second argument is definite, other argument becomes equal 
(define (eq-result sub-aim)
;  (cout "eq-result " sub-aim nl)
  (map (lambda (x) 
         (make-vector 2 x)) 
    (if (const? (second sub-aim)) (second sub-aim) (third sub-aim))))
;(cout (eq-result '(eq (1 2 3) Y)))
;=> (#(1 1) #(2 2) #(3 3))
;(trace eq-result)

; internal-preds - list of pairs: (name of pred . function implements calculation)
(define internal-preds `((eq ,eq-result)))
;(cout ((cadr (car internal-preds)) '(eq (1 2 3) Y)) nl)
;=> (#(1 1) #(2 2) #(3 3))

; empty-pred is a empty predicate or empty table
(define empty-pred '())

(define (value->number v)
  (define iv (if (string? v) (string->number v) v)) 
  (if iv iv 
    (error "Semantic error"  "Value must be integer type " iv)))
    
